''
''
''  Elev8 QuadRotor - Full Stability
''       -- By Jason Dorie --               
''               
''
'' This code assumes a HoverFly Open board is used, meaning:
''
'' The RC Receiver pins are:
''    Throttle:      P9
''    Aileron:       P8
''    Elevator:      P10
''    Rudder:        P11
''    Gain:          P13
''    Extra channel: P12
''
''
'' An ITG-3200 with pullups connected to SDA / SCL
''    INT: P14  (unused)
''    SDA: P15
''    SCL: P16
''
''
'' Electronic speed controls (ESCs) connected to pins:
''
''   Top row:     P30    P31    P0    P1
''   Bottom row:  P27    P26    P3    P2   (use the bottom row for X or + config)
''
'' RGB LED connected to P4, P5, P6
''
'' Speaker connected to P7



CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  '' Motor pins for '+' shaped frame config
  OUT_F  = 3                    ' Front ESC
  OUT_R  = 5                    ' Right ESC
  OUT_B  = 2                    ' Back ESC
  OUT_L  = 4                    ' Left ESC

  '' Motor pins for 'X' shaped frame config
  OUT_FL = 4                    ' Front Left ESC                       
  OUT_FR = 3                    ' Front Right ESC
  OUT_BR = 5                    ' Back Right ESC                        
  OUT_BL = 2                    ' Back Left ESC


  '' Indices for R/C receiver inputs 
  IN_ELEV = 2
  IN_THRO = 1
  IN_AILE = 0
  IN_RUDD = 3
  IN_AUX  = 4
  IN_GAIN = 5


  'Pins for various LED colors
  Red   = %010
  Green = %001
  Blue  = %100

  Yellow= %011
  Cyan  = %101
  White = %111

  'Constants for digital R/C filtering
  Filter_Heavy =    $2AAAAAAA         '0.1666 ** 2^32  (RC = 0.0200)   
  Filter_Moderate = $2F9DB22D         '0.1860 ** 2^32  (RC = 0.0175)   
  Filter_Light =    $4923A29C         '0.2857 ** 2^32  (RC = 0.0100)   



OBJ
  Gyro      : "ITG-3200-Open-pasm.spin"                 ' 1 cog
  RC        : "RC_Receiver-Open.spin"                   ' 1 cog                       
  ESC       : "Servo8Fast.spin"                         ' 1 cog
  Dbg       : "FullDuplexSerial.spin"                   ' 1 cog

  RollPID   : "IntPID-scaled.spin"
  PitchPID  : "IntPID-scaled.spin"
  YawPID    : "IntPID-scaled.spin"

  Eeprom    : "Basic_I2C_Driver.spin"
  Freq      : "Synth.spin"
  

VAR
  long Output[8]                                        ' Temp output flight controls array, copied into ESC when finished                        
  long counter
  long Thro, Aile, Elev, Rudd, Gain                     ' Receiver input values                        
  long iAile, iElev, iRudd                              ' Integrated input values (just summed over time)

  long DesiredRoll, DesiredPitch, DesiredYaw
  long RollOut, PitchOut, YawOut                        ' Output values from PIDs
  long ThroMix                                          ' Scale factor for flight outputs based on throttle value
  long LastTime, CurTime                                ' Loop timers

  long GXLP, GYLP, GZLP                                 ' Temp gyro readings for low pass filter

  long Gx, Gy, Gz                                       ' Gyro readings (instant)
  long iGx, iGy, iGz                                    ' Gyro readings (integrated)

  long DoIntegrate

  long FlightEnabled
  long EnableStep

  long DataMode                                         ' What data to send to the host
  long rxResult                        



PUB Main                              

  LED( Red )

  FlightEnabled := FALSE        'Disable the flight / motor loop
  DIRA[7] := 1                  'Speaker is an output

  'Pull the CLKIN pin low on the gyro
  DIRA[17] := 1
  OUTA[17] := 0  

  Dbg.Start( 31, 30, 0, 115200 )
  Gyro.Start( 16, 15 )          'Start the gyro reader object 

  DataMode := 0                 'Send adjusted gyro readings



  ' Initialize servo values to all neutral (zero for throttles)
  ' Delay values of 1500 = center, 1000 = min, 2000 = max

  if( ThrottleStartupTime == 0 )
    'Motors off on startup
    Output[0] := Output[1] := Output[2] := Output[3] := 1000
    Output[4] := Output[5] := Output[6] := Output[7] := 1000
  else
    'Motors to full throttle on startup for range setting    
    Output[0] := Output[1] := Output[2] := Output[3] := 2000
    Output[4] := Output[5] := Output[6] := Output[7] := 2000


  'Start the servo driver cog
  ESC.Init
  ESC.AddFastPin(0)
  ESC.AddFastPin(1)
  ESC.AddFastPin(2)
  ESC.AddFastPin(3)
  ESC.AddFastPin(26)
  ESC.AddFastPin(27)  
  'ESC.AddFastPin(30)           'Removed because they interfere with serial debug output
  'ESC.AddFastPin(31)  

  ESC.Set(  0, Output[0] )
  ESC.Set(  1, Output[1] )
  ESC.Set(  2, Output[2] )
  ESC.Set(  3, Output[3] )
  ESC.Set(  4, Output[4] )
  ESC.Set(  5, Output[5] )
  ESC.Set(  6, Output[6] )
  ESC.Set(  7, Output[7] )

  ESC.Start


  'Motor startup max throttle setting, if specified
  counter := (ThrottleStartupTime * 80_000 )            'Convert value to milliseconds
  if( counter => 400 )
    waitcnt( counter + cnt )
    'Motors off
    Output[0] := Output[1] := Output[2] := Output[3] := 1000
    Output[4] := Output[5] := Output[6] := Output[7] := 1000
    
  ESC.Set(  0, Output[0] )
  ESC.Set(  1, Output[1] )
  ESC.Set(  2, Output[2] )
  ESC.Set(  3, Output[3] )
  ESC.Set(  4, Output[4] )
  ESC.Set(  5, Output[5] )
  ESC.Set(  6, Output[6] )
  ESC.Set(  7, Output[7] )
     

  'Initialize the RC input object
  RC.start  

  'Initialize the PID loop scales - These values are stored in the DAT section at the end of the file

  'These numbers are scaled by 2^32 - the PID code internally divides the result by 2^32 on output
  'The shifts in the Init calls are just to allow the use of more human-readable values.

  RollPID.Init(    PitchRoll_PGain,  PitchRoll_IGain,  PitchRoll_DGain )
  PitchPID.Init(   PitchRoll_PGain,  PitchRoll_IGain,  PitchRoll_DGain )
  YawPID.Init(     Yaw_PGain,        Yaw_IGain,        Yaw_DGain )                    

  RollPID.SetD2Gain(  PitchRoll_D2Gain )
  PitchPID.SetD2Gain( PitchRoll_D2Gain )
  'YawPID.SetD2Gain(   Yaw_D2Gain )


  'Should probably wait for the gyro to finish startup here, but it takes a while
  'for the ESCs to initialize anyway...

  'repeat until( Gyro.IsReady == 1 )

  'TestRC                       'Uncomment to test pin inputs from R/C receiver


  'Run the flight loop
  FlightLoop



PUB FlightLoop | t, PR_P, PR_D, Y_P, Y_D, AbsGain

  counter := 0

  DesiredRoll := DesiredPitch := DesiredYaw := 0

  EnableStep := 0               'Counter to know which section of enable/disable sequence we're in
  LED( Green )                  'Green = disarmed, friendly

  'Record the time as we enter the flight loop
  LastTime := cnt

  repeat

    rxResult := Dbg.rxCheck
    if( rxResult <> -1 )
      if( rxResult == $AA )
        DoConfig
        LastTime := cnt           'Reset the loop wait counter or we'll be waiting a while
       
      elseif( rxResult == $AB )
        DataMode := 0
        
      elseif( rxResult == $AC )
        DataMode := 1

     
    'Read the values from the remote control  
    Thro := RC.getrc(IN_THRO)
    Aile := RC.getrc(IN_AILE)
    Elev := RC.getrc(IN_ELEV)
    Rudd := RC.getrc(IN_RUDD)
    Gain := RC.getrc(IN_GAIN)


    'Small dead band around zero to help prevent control drift
    if( ||Aile < 4 )
      Aile := 0

    if( ||Elev < 4 )
      Elev := 0

    if( ||Rudd < 4 )
      Rudd := 0


    if( Gain > 0 )                                      'Gear switch forward?  (normal)
      'Standard results from the gyro
      Gx := -Gyro.GetRY                                 'note that the gyro is mounted with X = left/right, Y = front/back
      Gy := -Gyro.GetRX
      Gz := Gyro.GetRZ

      PR_P := PitchRoll_PGain * Gain
      PR_D := PitchRoll_DGain * Gain
      Y_P := Yaw_PGain * Gain
      Y_D := Yaw_DGain * Gain
      
    else
      'Low pass filtered numbers
      GXLP := -Gyro.GetRY                               'note that the gyro is mounted with X = left/right, Y = front/back
      GYLP := -Gyro.GetRX
      GZLP := Gyro.GetRZ

      Gx := Gx + Filter_Light ** (GXLP - Gx) 
      Gy := Gy + Filter_Light ** (GYLP - Gy)
      Gz := Gz + Filter_Light ** (GZLP - Gz)

      AbsGain := -Gain
      PR_P := PitchRoll_PGain * AbsGain
      PR_D := PitchRoll_DGain * AbsGain
      Y_P := Yaw_PGain * AbsGain
      Y_D := Yaw_DGain * AbsGain


    PitchPID.SetPIDGain( PR_P, 0, PR_D )       
    RollPID.SetPIDGain( PR_P, 0, PR_D )       
    YawPID.SetPIDGain( Y_P, 0, Y_D )       


    if( FlightEnabled == FALSE )

      'Left stick positioned in lower left?
      if( EnableStep == 0  AND  Thro < -380  AND  Rudd >  350 )
        EnableStep := 1
        Beep
        Led( Green | Red )        
        LastTime := cnt


      'Left stick positioned in lower right?
      if( EnableStep == 1  AND  Thro < -380  AND  Rudd < -350 )
        FlightEnabled := TRUE
        EnableStep := 0
        Beep2

        LED( Red )
        Gyro.FindZero

        LED( Blue )        
        BeepTune
        LastTime := cnt


    else


      'Left stick positioned in lower left?
      if( EnableStep == 0  AND  Thro < -380  AND  Rudd >  350 )
        EnableStep := 1
        Beep
        Led( Green | Red )        
        LastTime := cnt


      'Left stick positioned in lower left?
      if( EnableStep == 1  AND  Thro < -380  AND  Rudd < -350 )
        FlightEnabled := FALSE
        EnableStep := 0
        Beep2
        LED( Green )        
        LastTime := cnt


      'Integrate the user stick values - this is just summing the values over time    
      iAile += Aile
      iElev += Elev
      iRudd += Rudd

         
      'Also integrate the gyro readings - just accumulate them over time       
      iGx += Gx
      iGy += Gy
      iGz += Gz
       
       
      'Zero all targets when throttle is off - makes for more stable liftoff
      if( Thro < -360 )
        iGx := iAile := 0
        iGy := iElev := 0
        iGz := iRudd := 0
        DoIntegrate := FALSE
      else
        DoIntegrate := TRUE
       

      'Scale the integrated user control values by the control scale amount
      'This allows for variable sensitivity.  Higher values make it faster.
       
      DesiredRoll := iAile * PitchRollControlScale 
      DesiredPitch := iElev * PitchRollControlScale
      DesiredYaw := iRudd * YawControlScale
       
       
      'Run the PID loops to compute the correction factors for each axis
       
      RollOut := RollPID.Calculate( DesiredRoll , iGx , DoIntegrate )    
      PitchOut := PitchPID.Calculate( DesiredPitch , iGy , DoIntegrate )    
      YawOut := YawPID.Calculate( DesiredYaw , iGz , DoIntegrate )    

       
      'Throttle mix ramps up the axis corrections as the throttle is moved from zero to
      'about 1/4 throttle.  This prevents the motors from turning on when you're carrying
      'the thing around.  :)
       
      ThroMix := (Thro + 400) ~> 2                        ' Approx 0 - 256
      ThroMix <#= 64                                      ' Above 1/4 throttle, clamp it to 64
      ThroMix #>= 0
       
       
      'add 1500 to all Output values to make them 'servo friendly' again
      Thro += 1500
       
       
      if( FrameConfiguration == 0 )
        ' + configuration
        Output[OUT_F] := Thro + ((-PitchOut + YawOut) * ThroMix) ~> 6                          
        Output[OUT_R] := Thro + (( RollOut  - YawOut) * ThroMix) ~> 6  
        Output[OUT_B] := Thro + (( PitchOut + YawOut) * ThroMix) ~> 6
        Output[OUT_L] := Thro + ((-RollOut  - YawOut) * ThroMix) ~> 6
         
      else
        ' X configuration
        Output[OUT_FL] := Thro + ((-PitchOut - RollOut - YawOut) * ThroMix) ~> 6                          
        Output[OUT_FR] := Thro + ((-PitchOut + RollOut + YawOut) * ThroMix) ~> 6  
        Output[OUT_BL] := Thro + (( PitchOut - RollOut + YawOut) * ThroMix) ~> 6
        Output[OUT_BR] := Thro + (( PitchOut + RollOut - YawOut) * ThroMix) ~> 6
         
       
       
      'Clamp all the outputs to be in range       
      Output[0] := 1000 #> Output[0] <# 2000
      Output[1] := 1000 #> Output[1] <# 2000
      Output[2] := 1000 #> Output[2] <# 2000
      Output[3] := 1000 #> Output[3] <# 2000
      Output[4] := 1000 #> Output[4] <# 2000
      Output[5] := 1000 #> Output[5] <# 2000
      Output[6] := 1000 #> Output[6] <# 2000
      Output[7] := 1000 #> Output[7] <# 2000
       
       
      'Copy new Ouput array into servo values
      ESC.Set( 0, Output[0] )
      ESC.Set( 1, Output[1] )
      ESC.Set( 2, Output[2] )
      ESC.Set( 3, Output[3] )
      ESC.Set( 4, Output[4] )
      ESC.Set( 5, Output[5] )
      ESC.Set( 6, Output[6] )
      ESC.Set( 7, Output[7] )



    if( DataMode == 0 )
      'Send gyro readings at the end of frame
      dbg.tx( $7f )
      dbg.tx( $7f )
      dbg.tx( gx >> 8 )
      dbg.tx( gx )
      dbg.tx( gy >> 8 )
      dbg.tx( gy )
      dbg.tx( gz >> 8 )
      dbg.tx( gz )
      
    elseif( DataMode == 1 )
      counter++
      if( counter == 5 )
        counter := 0
        dbg.tx( $7f )
        dbg.tx( $73 )
        t := Gyro.GetTemp
        dbg.tx( t >> 8 )
        dbg.tx( t )
        t := Gyro.GetRawX
        dbg.tx( t >> 8 )
        dbg.tx( t )
        t := Gyro.GetRawY
        dbg.tx( t >> 8 )
        dbg.tx( t )
        t := Gyro.GetRawZ
        dbg.tx( t >> 8 )
        dbg.tx( t )


    waitcnt( constant(80_000_000 / 250) + LastTime )
    LastTime += constant(80_000_000 / 250)



pub DoConfig | start, end, len, val, i, timer

  'Check the next RX byte for either $A5 or $5A
  waitcnt( 2_000 + cnt )

  start := @ConfigStart
  end := @ConfigEnd
  len := end - start

  end -= 1                      'Makes the loops simpler as they include the last value
   
  val := Dbg.rxCheck
  if( val == $A5 )              '$A5 = Query config

    Dbg.tx( $7f )
    Dbg.tx( $7e )

    Dbg.tx( len>>2 )

    repeat i from start to end
      Dbg.tx( byte[i] )


  elseif( val == $5A )          '$5A = Set config

    if( len == 64 )
      'Get the new configuration data from the serial port and apply it to memory
      repeat i from start to end
        byte[i] := Dbg.rx
       
      Dbg.tx( $7F )               'Send ACK of received config
      Dbg.tx( $70 )
       
      'Write it to EEPROM

      repeat i from start to end step 32
       
        if Eeprom.WritePage(Eeprom#BootPin, Eeprom#EEPROM, i, i, 32)
          Dbg.tx( $7F )
          Dbg.tx( $79 )       'NAK - error
          waitcnt( 8_000_000 + cnt )
          abort ' an error occured during the write
           
        Dbg.tx( $7F )
        Dbg.tx( $71 )
         
        timer := cnt ' prepare to check for a timeout
        repeat while Eeprom.WriteWait(Eeprom#BootPin, Eeprom#EEPROM, i)
          if cnt - timer > clkfreq / 10
            Dbg.tx( $7F )
            Dbg.tx( $79 )     'NAK - error
            waitcnt( 8_000_000 + cnt )
            abort ' waited more than a 1/10 second for the write to finish
         
      Dbg.tx( $7F )
      Dbg.tx( $72 )
      waitcnt( 8_000_000 + cnt )
         
      'Reset the prop
      reboot
       


PUB BeepHz( Hz , Delay )

    Freq.Synth("A", 7, Hz )
    waitcnt( Delay*80_000 + cnt )
    Freq.Synth( "A", 7, 0 )        


PUB BeepTune

  BeepHz( 1174, 250 )           'D5
  BeepHz( 1318, 250 )           'E5
  BeepHz( 1046, 250 )           'C5
  BeepHz(  522, 250 )           'C4
  BeepHz(  784, 500 )           'G4



PUB Beep
  BeepHz( 1318 , 125 )


PUB Beep2
  Beep
  waitcnt( 20_000_000 + cnt ) 
  Beep



PUB LED( bits )
  DIRA[4..6] := bits



PUB TestRC | index

  index := 0
  repeat
    dbg.tx(1)                    'Home cursor
    dbg.tx(14)                  'Position in x
    dbg.tx( index*6 )

    dbg.decd( rc.GetRC(index) , 6 )

    index++
    index //= 6
    
    waitcnt( 2_000_000 + cnt )      



DAT

ConfigStart

GyroOrientation         long  0                 ' 0 = normal, 1 = rotated 90 CCW, 2 = rotated 90 CW
FrameConfiguration      long  1                 ' 0 = +, 1 = x 

PitchRollControlScale   long  6                 ' 16 (insane), 8 (rapid), 4 (agile), 2 (slower), 1 (beginner)  (or anywhere in between)
YawControlScale         long  6


'The PID gains have all been scaled down by 200, assuming you'll set your gain to that on the TX

PitchRoll_PGain         long  $1EB8             '$18_0000 / 200
PitchRoll_IGain         long  $0  
PitchRoll_DGain         long  $28F5C2           '$2000_0000 / 200
PitchRoll_D2Gain        long  0

Yaw_PGain               long  $3D70             '$30_0000 / 200
Yaw_IGain               long  $0  
Yaw_DGain               long  $333333           '$2800_0000 / 200
       
ThrottleStartupTime     long  0                 'Hold time in mSec (0 = no throttle range set on startup)
DummyVar                long  0,0,0,0           'Padding, to keep the EEPROM write at 64 bytes

' The previous data is 16 longs, or 64 bytes total
ConfigEnd